/*
	BMP Deinterlacer -- Deinterlaces BMP image files
	Copyright (C) 2001 Donald A. Graft

	Field-mode combing detection conceived by Gunnar Thalin.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	Donald Graft
	neuron2@home.com.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <windows.h>

void main(int argc, char *argv[])
{
	FILE *fp;
	FILE *ofp;
	unsigned char *bitmap, *obitmap, *bits, *obits, *moving, *fmoving;
	long bwidth, bheight, bpitch;
	unsigned int depth;
	long size;
	int x, y;
	unsigned char *src, *srcminus, *srcplus, *dst;
	long gr, gg, gb, gt;
	long threshold = 20 * 20;
	char c;
	int first = 1;
	int i = 1;
	int show_combing = 0;
	unsigned char *mov, *fmov;
	int disable_denoise = 0;
	char *p, path[MAX_PATH], opath[MAX_PATH], tmp[5];
	int batch = 0;
	WIN32_FIND_DATA data;
	HANDLE handle;
	char *batchdir;
	void usage(void);

	/* Process command line arguments. */
	argc--;
	if (argc < 1)
	{
		usage();
		return;
	}
	while (argc != 0 && argv[i][0] == '-')
	{
		switch (argv[i][1])
		{
			case 'h':
			case '?':
				/* Help screen */
				usage();
				return;
			case 'b':
				/* Batch mode */
				batch = 1;
				argc--;
				i++;
				break;
			case 'd':
				/* Disable combing map denoising */
				disable_denoise = 1;
				argc--;
				i++;
				break;
			case 'c':
				/* Show combed areas */
				show_combing = 1;
				argc--;
				i++;
				break;
			case 't':
				/* Combing detection threshold */
				argc -= 2;
				if (argc == 0)
				{
					usage();
					return;
				}
				++i;
				if (*argv[i] < '0' || *argv[i] > '9')
				{
					printf("Bad threshold\n");
					return;
				}
				threshold = atol(argv[i]);
				threshold *= threshold;
				++i;
				break;
			default:
				usage();
				return;
		}
	}
	if ((batch && argc > 1) || (!batch && (argc < 1 || argc > 2)))
	{
		usage();
		return;
	}

	/* Load and parse the input BMP file(s). Allocate needed memory. */
	if (batch)
	{
		if (argc == 0) batchdir = ".";
		else batchdir = argv[i];
		strcpy(path, batchdir);
		if (path[strlen(path)-1] != '\\')
			strcat(path, "\\");
		strcat(path, "*.bmp");
		handle = FindFirstFile(path, &data);
		if (handle == INVALID_HANDLE_VALUE)
		{
			printf("No BMP files found\n");
			return;
		}
		strcpy(path, batchdir);
		if (path[strlen(path)-1] != '\\')
    		strcat(path, "\\");
		strcat(path, data.cFileName);
//		printf("%s\n", path);
		if ((fp = fopen(path, "rb")) == NULL)
		{
			printf("Cannot open input file %s\n", path);
			return;
		}
	}
	else
	{
		strcpy(path, argv[i]);
		if ((fp = fopen(path, "rb")) == NULL)
		{
			printf("Cannot open input file %s\n", path);
			return;
		}
	}

	while(1)
	{
		if (batch && !first)
		{
			if (FindNextFile(handle, &data))
			{
				strcpy(path, batchdir);
				if (path[strlen(path)-1] != '\\')
    				strcat(path, "\\");
				strcat(path, data.cFileName);
//				printf("%s\n", path);
				if ((fp = fopen(path, "rb")) == NULL)
				{
					printf("Cannot open input file %s\n", path);
					return;
				}
			}
			else
			{
				FindClose(handle);
				free(bitmap);
				free(obitmap);
				free(moving);
				free(fmoving);
				return;
			}
		}
		first = 0;

		fread(tmp, 1, 2, fp);
		if (tmp[0] != 'B' || tmp[1] != 'M')
		{
			printf("Not a BMP file\n");
			return;
		}
		fread(&size, 1, 4, fp);
		fclose(fp);
		bitmap = (unsigned char *) malloc((size_t) size);
		obitmap = (unsigned char *) malloc((size_t) size);
		if (bitmap == NULL || obitmap == NULL)
		{
			printf("Cannot malloc\n", size);
			return;
		}
		if ((fp = fopen(path, "rb")) == NULL)
			return;
		fread(bitmap, 1, size, fp);
		fclose(fp);
		memcpy(obitmap, bitmap, size);
		bits = bitmap + *((long *)(bitmap + 10));
		obits = obitmap + *((long *)(obitmap + 10));
		bwidth = *((long *)(bitmap + 18));
		bheight = *((long *)(bitmap + 22));
		bpitch = 3 * bwidth;
		if (bpitch & 0x3)
		{
			bpitch = (bpitch & 0xfffffffc) + 4;
		}
		depth = *((unsigned int *)(bitmap + 28));
		if (depth != 24)
		{
			printf("BP file must have 24-bit color depth\n");
			return;
		}
		moving = (unsigned char *) malloc((size_t) bheight * bwidth);
		fmoving = (unsigned char *) malloc((size_t) bheight * bwidth);
		if (moving == NULL || fmoving == NULL)
		{
			printf("Cannot malloc\n", size);
			return;
		}
		memset(moving, 0, bheight * bwidth);
		memset(fmoving, 0, bheight * bwidth);

		/* Calculate the combing map. */
		src = bits;
		srcminus = src - bpitch;
		srcplus = src + bpitch;
		for (y = 0; y < bheight; y++)
		{
			for (x = 0; x < bpitch; x+=3)
			{
				if ((y & 1) && (y < bheight - 1))
				{
					/* Combing detection concept by Gunnar Thalin. */
					gr = 3 * ((srcminus[x+2] - src[x+2]) *
								 (srcplus[x+2]  - src[x+2]));
					gg = 6 * ((srcminus[x+1] - src[x+1]) *
								 (srcplus[x+1] - src[x+1]));
					gb =     ((srcminus[x] - src[x]) *
								 (srcplus[x]  - src[x]));
					gt = (gr + gg + gb);

					if (gt > threshold)
					{
						moving[y * bwidth + x/3] = 1;
					}
					/* End concept by Gunnar Thalin. */
				}
			}
			src += bpitch;
			srcminus += bpitch;
			srcplus += bpitch;
		}

		/* If enabled, denoise the combing map. */
		if (!disable_denoise)
		{
			int xlo, xhi, ylo, yhi;
			int u, v;
			int N = 3;
			int Nover2 = N/2;
			int sum;
			unsigned char *m;
			int w = bwidth;
			int h = bheight;
			int wminus1 = w - 1;
			int hminus1 = h - 1;

			/* Remove single points from the map. This code is
			   more general than needed for that to allow for trying
			   different denoising approaches. */
			fmov = fmoving;
			for (y = 0; y < h; y++)
			{
				for (x = 0; x < w; x++)
				{
					if (!((moving + y * w)[x]))
					{
						fmov[x] = 0;	
						continue;
					}
					xlo = x - Nover2; if (xlo < 0) xlo = 0;
					xhi = x + Nover2; if (xhi >= w) xhi = wminus1;
					ylo = y - Nover2; if (ylo < 0) ylo = 0;
					yhi = y + Nover2; if (yhi >= h) yhi = hminus1;
					m = moving + ylo * w;
					sum = 0;
					for (u = ylo; u <= yhi; u++)
					{
						for (v = xlo; v <= xhi; v++)
						{
							sum += m[v];
						}
						m += w;
					}
					if (sum > 1)
						fmov[x] = 1;
					else
						fmov[x] = 0;
				}
				fmov += w;
			}
		}

		/* Render. */
		dst = obits;
		srcminus = bits - bpitch;
		srcplus = bits + bpitch;
		if (!disable_denoise) mov = fmoving;
		else mov = moving;
		for (y = 0; y < bheight; y++)
		{
			for (x = 0; x < bpitch; x+=3)
			{
				if ((y & 1) && (y < bheight - 1) && mov[y * bwidth + x/3])
				{
					if (show_combing)
					{
						dst[x]   = 0;
						dst[x+1] = 0;
						dst[x+2] = 255;
					}
					else
					{
						dst[x]   = (srcminus[x]   + srcplus[x])   >> 1;
						dst[x+1] = (srcminus[x+1] + srcplus[x+1]) >> 1;
						dst[x+2] = (srcminus[x+2] + srcplus[x+2]) >> 1;
					}
				}
			}
			srcminus += bpitch;
			srcplus += bpitch;
			dst += bpitch;
		}

		/* Write output BMP file. */ 
		if (argc == 2)
		{
			/* Output file name given; use it. */
			printf("Creating %s\n", argv[++i]);
			if ((ofp = fopen(argv[i], "wb")) == NULL)
			{
				printf("Cannot open output file %s\n", argv[i]);
				return;
			}
		}
		else
		{
			/* Output file name not given or batch mode; create an output
			   directory and use the input file name. */
			/* Get current directory and create directory "deint"
			   below it. */
			getcwd(opath, MAX_PATH);
			if (opath[strlen(opath)-1] != '\\')
	    		strcat(opath, "\\");
			strcat(opath, "deint");
			mkdir(opath);
			/* Create path to output file and open it. */
			strcat(opath, "\\");
			p = path;
			p += strlen(path);
			while (--p >= path && *p != '\\');
			p++;
			strcat(opath, p);
			printf("Creating %s\n", opath);
			if ((ofp = fopen(opath, "wb")) == NULL)
			{
				printf("Cannot open output file %s\n", opath);
				return;
			}
		}
		fwrite(obitmap, 1, size, ofp);
		fclose(ofp);

		free(bitmap);
		free(obitmap);
		free(moving);
		free(fmoving);

		if (!batch) break;
	}
}

void usage(void)
{
	printf("\nBMP Deinterlacer (1.3) by Donald Graft, neuron2@home.com\n\n");
	printf("Usage 1: bmpdeint [options] infile [outfile]\n\n");
	printf("If outfile is omitted, the output will be placed in a directory\n");
	printf("called deint (created if necessary) under the current directory.\n");
	printf("The output file will have the same name as the input file.\n\n");
	printf("Usage 2: bmpdeint [options] -b dir\n\n");
	printf("All BMP files in the specified directory will be processed\n");
	printf("and the resulting processed BMP files will be placed in a\n");
	printf("directory called deint (created if necessary) under the current\n");
	printf("directory. The output files will have the same names as the\n");
	printf("corresponding input files. To specify the current directory,\n");
	printf("use \"bmpdeint -b .\" or \"bmpdeint -b\".\n\n");
	printf("	-h	print help\n");
	printf("	-?	print help\n");
	printf("	-c	show combing areas\n");
	printf("	-d	disable denoising\n");
	printf("	-t n	combing threshold is n (default 20)\n");
}
